from .._utils import init_shared_object, setup_app


def main():
    # config = init_shared_object(specific_config_path)
    # setup_app(config)
    raise SystemExit("Gui version hasn't been implemented yet")
